/*
 * global.c - any/all global data items etc should be in this file
 *
 *  $Id: global.c,v 1.16 1998/08/08 20:08:09 johns Exp $
 *
 */

int parinitted = 0;                 /* used by parallel startup code     */

