/*
 * global.h - any/all global data items etc should be in this file
 *
 *  $Id: global.h,v 1.16 2011/02/05 08:10:11 johns Exp $
 *
 */

extern int parinitted;  /**< global initialization state variable */

